//
//  It is file 'archive.cc'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
#include "stdinc.hpp"
#include "timestamp.hpp"
#include "byteorder.hpp"
#include "archdr.hpp"
#include "machine.hpp"
#include "archive.hpp"

Archive::Archive()
{
incomplete = 0 ;
name	   = NULL ;
err_msg    = NULL ;
handle	   = -1 ;
}

Archive::~Archive()
{
if( incomplete )
     Delete() ;
else Close() ;
free( name ) ;
name = NULL ;
}

void
Archive::Perror( const char * const msg ) const
{
if( name != NULL )
    cerr << name << ": " ;
if( msg != NULL )
    cerr << msg << ": " ;
if( err_msg != NULL )
    cerr << err_msg ;
cerr << "\n" ;
}


int
Archive::Open( const char * const n )
{
name	  = strdup( n ) ;
if( ( handle = open( name, O_RDONLY | O_BINARY ) ) == -1 ){
    err_msg = strerror( errno ) ;
    return -1 ;
    }

    char	header[ sizeof ArchiveSignature - 1 ] ;

err_msg = "File is not an archive" ;
if( read( handle, header, sizeof header ) == -1 )
    return -1 ;
if( memcmp( header, ArchiveSignature, sizeof header ) != 0 )
    return -1 ;
err_msg = NULL ;
return 0 ;
}

int
Archive::Create( const char * const n )
{
incomplete = 1 ;
name	   = new char[ strlen( n ) + 1 + 4 ] ;
strcpy( name, n ) ;
    {
	char	   *p = strrchr( name, '.' ) ;

    if( p == NULL ) p = name + strlen( name ) ;
    strcpy( p, ".$" ) ;

	int	    i ;

    for( i = 0, p += 2 ; i < 256 ; i++ ){
	itoa( i, p, 16 ) ;
	if( access( name, 0 ) == -1 )
	    break ;
	}
    if( i >= 256 ){
	err_msg = "Can't generate name for temporary archive" ;
	return -1 ;
	}
    }
if( ( handle = open( name, O_RDWR | O_CREAT | O_TRUNC | O_BINARY, S_IREAD | S_IWRITE | S_IFREG ) ) == -1 ){
    err_msg = strerror( errno ) ;
    return -1 ;
    }
if( write( handle, ArchiveSignature, sizeof ArchiveSignature - 1 ) != sizeof ArchiveSignature - 1 ){
    err_msg = strerror( errno ) ;
    return -1 ;
    }
return 0 ;
}

int
Archive::Close( void )
{
if( handle != -1 )
    close( handle ) ;
handle	   = -1 ;
incomplete = 0 ;
return 0 ;
}

int
Archive::Delete( void )
{
Close() ;
if( name != NULL && unlink( name ) != 0 ){
    err_msg = strerror( errno ) ;
    return -1 ;
    }
return 0 ;
}

int
Archive::Backup()
{
if( access( name, 0 ) == -1 ){
    delete name ;
    name = NULL ;
    return 0 ;
    }
char *new_name = new char [ strlen( name ) + 1 + 3 ] ;
strcpy( new_name, name ) ;
char *p = strrchr( new_name, '.' ) ;
if( p == NULL )
    p = new_name + strlen( new_name ) ;
strcpy( p, ".$bk" ) ;
unlink( new_name ) ;
int code = Rename( new_name ) ;
delete new_name ;
return code ;
}

int
Archive::Rename( const char * const new_name )
{
if( handle != -1 ){
    close( handle ) ;
    handle = -1 ;
    }
if( rename( name, new_name ) != 0 ){
    err_msg = strerror( errno ) ;
    return -1 ;
    }
delete name ;
name = strdup( new_name ) ;
return 0 ;
}

int
Archive::Eof( void ) const
{
if( handle == -1 )
    return 1 ;
if( tell( handle ) >= filelength( handle ) )
    return 1 ;
return 0 ;
}

char * const
Archive::GetName( void ) const
{
return name ;
}

void
Archive::FetchBlockHeader( UncomprHdr & hdr )
{
if( read( handle, &hdr, SizeOfUncomprHdr ) != SizeOfUncomprHdr ){
    err_msg = strerror( errno ) ;
    Perror( "Archive::FetchBlockHeader" ) ;
    exit( -1 ) ;
    }
ToNative( &hdr, hdr.Layout ) ;
}

void
Archive::NextEntry()
{
    long	pos = tell( handle ) ;
    UncomprHdr	hdr ;

FetchBlockHeader( hdr ) ;
if( lseek( handle, pos + hdr.BlockSize, SEEK_SET ) > tell( handle ) ){
    err_msg = "Archive file too short" ;
    Perror( "Archive::NextEntry" ) ;
    exit( -1 ) ;
    }
}

void
Archive::FillFileInfo( FileInfo & info )
{
    CompletionCode  err = DecompressFile( handle, &info, GetInfo ) ;

if( err != Completed ){
    err_msg = MachineStrerror( err ) ;
    Perror( "Archive::FillFileInfo" ) ;
    exit( -1 ) ;
    }
}

void
Archive::CopyCurrentFragment( class Archive & in )
{
    long	in_pos	= tell( in.handle ) ;
    long	out_pos = tell( handle ) ;
    UncomprHdr	in_hdr ;
    char	buf[ BlockCopyBuffer ] ;

in.FetchBlockHeader( in_hdr ) ;
ToIntel( &in_hdr, in_hdr.Layout ) ;
if( write( handle, &in_hdr, SizeOfUncomprHdr ) != SizeOfUncomprHdr ){
    err_msg = strerror( errno ) ;
    Perror( "Archive::CopyCurrentFragment" ) ;
    exit( -1 ) ;
    }
ToNative( &in_hdr, in_hdr.Layout ) ;

    unsigned	step ;
    long	size = in_hdr.BlockSize - SizeOfUncomprHdr ;

//cerr << "Size of block to copy " << size << "\n" ;

for( long offset = 0 ; offset < size ; offset += step ){
    if( size - offset >= BlockCopyBuffer )
	 step = BlockCopyBuffer ;
    else step = (unsigned)( size - offset ) ;
    if( read( in.handle, buf, step ) != step ){
	in.err_msg = strerror( errno ) ;
	in.Perror() ;
	exit( -1 ) ;
	}
    if( write( handle, buf, step ) != step ){
	err_msg = strerror( errno ) ;
	Perror() ;
	exit( -1 ) ;
	}
    }
lseek( in.handle, in_pos, SEEK_SET ) ;
lseek( handle, out_pos, SEEK_SET ) ;
}

void
Archive::CompressFile( char * const name, int order, unsigned long memory )
{
CompletionCode	err = ::CompressFile( handle, name, order, memory ) ;
if( err != Completed ){
    err_msg = MachineStrerror( err ) ;
    Perror() ;
    exit( -1 ) ;
    }
}

int
Archive::ExtractCurrentFragment()
{
    FileInfo	dummy ;

CompletionCode	err = DecompressFile( handle, &dummy ) ;
if( err != Completed )
    err_msg = MachineStrerror( err ) ;
switch( err ){
    default:
	Perror() ;
	exit( -1 ) ;
    case FileNotFound:
    case BadCRC:
	return -1 ;
    case Completed:
	return 0 ;
    }
}

int
Archive::TestCurrentFragment()
{
    FileInfo	dummy ;

CompletionCode	err = DecompressFile( handle, &dummy, TestFragment ) ;
if( err != Completed )
    err_msg = MachineStrerror( err ) ;
switch( err ){
    default:
	Perror() ;
	exit( -1 ) ;
    case BadCRC:
	return -1 ;
    case Completed:
	return 0 ;
    }
}
