//
//  It is file 'memory.hpp'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//
//  Here is simple memory alllocator with small memory overhead. I was
//  forced to write it when I had seen what default memory allocator
//  from BC 3.1 wastes as much memory as it gives to me when I use relatively
//  small allocation blocks (under 32 bytes).
//
//  Memory allocator structure is a bit weird (it is the most
//  hardware-dependent part of all the program. It is a result
//  of high memory requerements of compression algorithm, which
//  could easily allocate couple of megabytes of memory in
//  8-byte chunks... I just can't cope with memory overhead of
//  default memory allocator.
//
//  Chunks of two types could be allocated with this class: fixed-size
//  unmovable headers (which corresponds to Context structures in compression
//  program) and movable fragments of variable size (Stats and Links arrays).
//  Fixed-size fragments are allocated starting from the top of available
//  memory, variable-sized - starting from the bottom. Then fragments are
//  released by client, they are placed to the corresponding free list. Upon
//  allocation, fragment of _exactly_that_ size is extracted from free list.
//  if such fragment do not exist, necessary fragment is removed from free
//  store area. When this is exhausted (either by fixed-size or variable-size
//  allocation), garbage collection is performed on the variable-size area.
//  Fixed-block area could only grow...
//
#ifdef MSDOS
#include <alloc.h>
#endif

//
//  Change the following definition to 'unsigned long', if your system
//  can't cope with 16-bit aligned 32-bit quantities. It will increase
//  allocator's memory overhead, though...
//
#ifdef ALIGNED_MEMORY_ACCESS
typedef unsigned long 	    BLKSIZE ;
#else
typedef unsigned short	    BLKSIZE ;
#endif

#include "pool.hpp"

const BLKSIZE	BlockUsed		= 1 ;

struct BlockHeader {
    union {
	struct BlockHeader *next ;		   // Free blocks only!
	void		  **reloc_ptr ;
	} ;
    BLKSIZE		    size ;
    BLKSIZE		    Size( void )    { return size & ~BlockUsed ; }
    Boolean		    Used( void )    { return size & BlockUsed ? True : False ; }
    struct BlockHeader *   &NextFree( void ){ return next ; }
    } ;

//
//  Necessary to beat silly alignment practiced by some compilers...
//
const	int	SizeOfBlockHeader = sizeof( void * ) + sizeof( BLKSIZE ) ;

struct FixedHeader {
    struct FixedHeader	   *next ;
    } ;

const int	MinAllocedBlock 	= SizeOfBlockHeader + sizeof( BLKSIZE ) ;
const int	MaxAllocedBlock 	= SizeOfBlockHeader + 1024 ;
const int	AllocationGranularity	= sizeof( BLKSIZE ) ;
const int	FreeListEntries 	= 1 + ( MaxAllocedBlock - MinAllocedBlock ) / AllocationGranularity ;

class MemoryAllocator : MemoryPool {
    unsigned long   block_size ;
    unsigned long   memory_free ;
    unsigned long   min_memory_free ;
    int 	    fixed_block_size ;
    BlockHeader    *variable_free_lists[ FreeListEntries ] ;
    FixedHeader    *fixed_free_list ;

    void	    GarbageCollection( void ) ;
    BlockHeader *  &GetVariableSlot( int &size ) ;
    void	    ClearFreeSlots() ;
    public:
		    MemoryAllocator( unsigned long, int fixed_block_size ) ;
		   ~MemoryAllocator( void ) ;
    unsigned long   MemoryUsed( void ) ;
    unsigned long   MemoryFree( void ) ;
    unsigned long   MaxMemoryUsed( void ) ;
    void *	    AllocateFixed( void ) ;
    void *	    AllocateMoveable( int size, void **ptr_loc ) ;
    void	    FreeFixed( void * ) ;
    void	    FreeMoveable( void * ) ;
    } ;

inline unsigned long
MemoryAllocator::MemoryUsed( void )
{
return block_size - memory_free ;
}

inline unsigned long
MemoryAllocator::MemoryFree( void )
{
return memory_free ;
}

inline unsigned long
MemoryAllocator::MaxMemoryUsed( void )
{
GarbageCollection() ;	// This is to produce accurate figures
			// on segmented architecture when memory
			// flush was never attempted yet.
return block_size - min_memory_free ;
}

inline void
MemoryAllocator::ClearFreeSlots()
{
for( int i = 0 ; i < FreeListEntries ; i++ )
    variable_free_lists[ i ] = NULL ;
}

inline BlockHeader * &
MemoryAllocator::GetVariableSlot( int &size )
{
size += SizeOfBlockHeader ;
#ifdef DEBUG
if( size > MaxAllocedBlock ){
    cerr << "MemoryAllocator::GetVariableSlot: block size " << size << " out of valid range\n" ;
    exit( -1 ) ;
    }
#endif
if( size < MinAllocedBlock )
    size = MinAllocedBlock ;
size	 = ( (size + (AllocationGranularity-1)) / AllocationGranularity ) * AllocationGranularity ;
return variable_free_lists[ (size - MinAllocedBlock) / AllocationGranularity ] ;
}

#ifdef GEN_HUGE_FUNCS
HugeInline
MemoryAllocator::MemoryAllocator( unsigned long size, int fixed_block ) :
    MemoryPool( size )
{
fixed_block_size  = fixed_block ;
min_memory_free   =
memory_free	  =
block_size	  = GetFreeSpace() ;
fixed_free_list   = NULL ;
ClearFreeSlots() ;
}
#endif

inline
MemoryAllocator::~MemoryAllocator( void )
{
}

inline void *
MemoryAllocator::AllocateFixed( void )
{
    void       *block ;

if( fixed_free_list != NULL ){
    block	    = (void *)fixed_free_list ;
    fixed_free_list = fixed_free_list->next ;
    #ifdef REPORT_ALLOCS
    cout << "Fixed block allocated from free list " ;
    #endif
    }
else {
    //
    //	No blocks in free list.
    //
    if( ( block = TopAlloc( fixed_block_size ) ) == NULL ){
	GarbageCollection() ;
	if( ( block = TopAlloc( fixed_block_size ) ) == NULL ){
	    cerr << "MemoryAllocator::AllocateFixed: free store exhausted\n" ;
	    exit( -1 ) ;
	    }
	}
    memory_free    -= fixed_block_size ;
    if( memory_free < min_memory_free )
	min_memory_free = memory_free ;
    #ifdef REPORT_ALLOCS
    cout << "Fixed block allocated from store " ;
    #endif
    }
#ifdef REPORT_ALLOCS
cout << "@ " << block << "\n" ;
#endif
return block ;
}

inline void
MemoryAllocator::FreeFixed( void *block )
{
if( block != NULL ){
	FixedHeader	*p = (FixedHeader *) block ;

    #ifdef REPORT_ALLOCS
    cout << "Fixed block released @" << p << "\n" ;
    #endif
    p->next	    = fixed_free_list ;
    fixed_free_list = p ;
//
//  This memory is not available for later variable-size reallocation,
//  so I will not include it in a free memory count.
//
//  memory_free += fixed_block_size ;
    }
}

inline void *
MemoryAllocator::AllocateMoveable( int size, void **p )
{
    BlockHeader *  &free_list = GetVariableSlot( size ) ;
    BlockHeader    *block ;

#ifdef REPORT_ALLOCS
cout << "Moveable block of size " << size ;
#endif
if( free_list != NULL ){
    block     = free_list ;
    free_list = free_list->next ;
    #ifdef REPORT_ALLOCS
    cout << " allocated from free list " ;
    #endif
    }
else {
    //
    //	No blocks in free list.
    //
    if( ( block = (BlockHeader *)BottomAlloc( size ) ) == NULL ){
	GarbageCollection() ;
	if( ( block = (BlockHeader *)BottomAlloc( size ) ) == NULL ){
	    cerr << "MemoryAllocator::AllocateMoveable: free store exhausted\n" ;
	    exit( -1 ) ;
	    }
	}
    #ifdef REPORT_ALLOCS
    cout << " allocated from store " ;
    #endif
    }
//
//  Now, when we have block we had longed for, do a few house-keeping
//  tricks and return memory block to caller.
//
block->size	  = (BLKSIZE) size | BlockUsed ;
block->reloc_ptr  = p ;
memory_free	 -= size ;
if( memory_free < min_memory_free )
    min_memory_free = memory_free ;
#ifdef REPORT_ALLOCS
cout << "@ " << block << "\n" ;
#endif
return (void *)( (unsigned char *)block + SizeOfBlockHeader ) ;
}

inline void
MemoryAllocator::FreeMoveable( void *p )
{
if( p != NULL ){
	BlockHeader    *block	  = (BlockHeader *) ( (unsigned char *)p - SizeOfBlockHeader ) ;
	int		size	  = block->Size() - SizeOfBlockHeader ;
	//  Attention! GetVariableSlot modifies it's parameter!
	BlockHeader *  &free_list = GetVariableSlot( size ) ;

    #ifdef DEBUG
    if( block->Size() != size || ! block->Used() ){
	cerr << "MemoryAllocator::FreeMoveable: memory allocation chain corrupted\n" ;
	exit( -1 ) ;
	}
    #endif
    block->size   = (BLKSIZE) size ;	// Clear busy bit
    block->next   = free_list ;
    free_list	  = block ;
    memory_free  += size ;
    #ifdef REPORT_ALLOCS
    cout << "Moveable block of size " << size << " released @ " << block << "\n" ;
    #endif
    }
}

#ifdef GEN_HUGE_FUNCS
HugeInline void
MemoryAllocator::GarbageCollection( void )
{
cerr << "*" << flush ;
ClearFreeSlots() ;

    MemoryPoolWalker	walk_from( *this ) ;

FlushBottomLine() ;	// Important! walk_from should be constructed _before_
			// FlushBottomLine is called!
while( (int) walk_from != 0 ){	// This stupid BC! It complains about ambiguity
				// of operator int with operator void * without explicit cnv.!
	BlockHeader    *from = (BlockHeader *) (void *) walk_from ;
	int		size = from->Size() ;

    walk_from	      += size ;
    if( from->Used() ) {
	    BlockHeader    *to	 = (BlockHeader *) BottomAlloc( size ) ;

	#ifdef DEBUG
	if( to == NULL ){
	    cerr << "MemoryAllocator::GarbageCollecton: Impossible out-of-memory condition\n" ;
	    exit( -1 ) ;
	    }
	if( from->Size() < MinAllocedBlock || from->Size() > MaxAllocedBlock ){
	    cerr << "MemoryAllocator::GarbageCollection: Heap corrupted\n" ;
	    exit( -1 ) ;
	    }
	if( *(from->reloc_ptr) != (void *)( (unsigned char *)from + SizeOfBlockHeader ) ){
	    cerr << "MemoryAllocator::GarbageCollection: Lost memory block detected\n" ;
	    exit( -1 ) ;
	    }
	#endif
	if( to != from ){
	    //
	    //	Move block to the new location
	    //
	    memcpy( to, from, size ) ;
	    *(to->reloc_ptr) = (void *)( (unsigned char *)to + SizeOfBlockHeader ) ;
	    }
	}
    }

memory_free = GetFreeSpace() ;
if( memory_free < min_memory_free )
    // This could happen only on segmented memory architectures...
    min_memory_free = memory_free ;

cerr << "\b \b" << flush ;
}
#endif
