//
//  It is file 'pool.hpp'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
class MemoryPool ;
class MemoryPoolWalker ;

#ifdef SEGMENTED_MEMORY
//const   unsigned    SegmentSize   = 31 * 1024 ;   // Lapin says it's a Good Thing...
const	unsigned    SegmentSize     = 63 * 1024 ;   // But with my value fragmentation is smaller ;)
const	unsigned    SegmentOverhead = 64 ;	    // System memory overhead, bytes per segment.
						    // This is mor or less arbitrary value...

class SegPtr {
    unsigned	    offset ;
    char * *	    segment ;
    public:
		    SegPtr(){} ;		// This is to make compiler happy with
						// enclosing classes. We'll always
						// re-initialize this class with a 'good'
						// constructor.
		    SegPtr( char * *, unsigned ) ;
		    operator void * () const ;
    SegPtr	    operator +	( int ) const ;
    SegPtr	    operator -	( int ) const ;
    SegPtr &	    operator += ( int ) ;
    int 	    operator == ( const SegPtr & ) const ;
    long	    operator -	( const SegPtr & ) const ;
    int 	    operator >	( const SegPtr & ) const ;
    int 	    operator <	( const SegPtr & ) const ;
    void	    NextSegment() ;
    void	    SetLimit() ;
    int 	    AtLimit() ;
    } ;

inline
SegPtr::SegPtr( char ** seg, unsigned off )
{
segment = seg ;
offset	= off ;
}

inline
SegPtr::operator void *() const
{
return (void *)( *segment + offset ) ;
}

inline int
SegPtr::operator == ( const SegPtr &two ) const
{
return segment == two.segment && offset == two.offset ;
}

inline void
SegPtr::NextSegment()
{
segment++ ;
offset = 0 ;
}

inline void
SegPtr::SetLimit()
{
((unsigned *)*segment)[-1] = offset ;
}

inline int
SegPtr::AtLimit()
{
return offset >= ((unsigned *)*segment)[-1] ;
}

inline SegPtr
SegPtr::operator + ( int diff ) const
{
if( diff < 0 ) return *this - (-diff) ;

    SegPtr	tmp( *this ) ;
    long	new_offset = long( tmp.offset ) + diff ;

if( new_offset >= SegmentSize ){
    tmp.segment++ ;
    if( new_offset > SegmentSize )
	 new_offset = diff ;
    else new_offset = 0 ;
    }
tmp.offset = unsigned( new_offset ) ;
return tmp ;
}

inline SegPtr
SegPtr::operator - ( int diff ) const
{
if( diff < 0 ) return *this + (-diff) ;

    SegPtr	tmp( *this ) ;
    long	new_offset = long( tmp.offset ) - diff ;

if( new_offset < 0 ){
    tmp.segment-- ;
    new_offset = SegmentSize - diff ;
    }
tmp.offset = unsigned( new_offset ) ;
return tmp ;
}

inline SegPtr &
SegPtr::operator += ( int diff )
{
return *this = *this + diff ;
}

inline int
SegPtr::operator > ( const SegPtr & two ) const
{
if( segment > two.segment ) return 1 ;
if( segment < two.segment ) return 0 ;
if( offset  > two.offset  ) return 1 ;
return 0 ;
}

inline int
SegPtr::operator < ( const SegPtr & two ) const
{
if( segment < two.segment ) return 1 ;
if( segment > two.segment ) return 0 ;
if( offset  < two.offset  ) return 1 ;
return 0 ;
}

inline long
SegPtr::operator - ( const SegPtr & two ) const
{
return long( segment - two.segment ) * long( SegmentSize ) + long( offset ) - long( two.offset ) ;
}

#else

typedef char *	    SegPtr ;

#endif

class MemoryPool {
    friend class    MemoryPoolWalker ;
#ifdef SEGMENTED_MEMORY
    int 	    segment_count ;
    char * *	    segments ;		// This points to start of segment + sizeof( unsigned ),
					// thus, ((unsigned *) *segments)[ -1 ] is the size
					// of used fragment of memory block...
#endif
    SegPtr	    memory_start ;
    SegPtr	    memory_top ;
    SegPtr	    free_store_bottom ;
    SegPtr	    free_store_top ;
    public:
		    MemoryPool( unsigned long size ) ;
		   ~MemoryPool() ;
    void *	    BottomAlloc( int size ) ;
    void *	    TopAlloc( int size ) ;
    void	    FlushBottomLine() ;
    unsigned long   GetFreeSpace() ;

    static unsigned long   AllocatorOverhead( unsigned long size ) ;
    } ;

class MemoryPoolWalker {
    friend class	MemoryPool ;
    SegPtr		memory_ptr ;
    SegPtr		memory_limit ;
    public:
			MemoryPoolWalker( MemoryPool & ) ;
    MemoryPoolWalker &	operator += ( int size ) ;
			operator void *() ;	    // Get current block address
			operator int() ;	    // Do we have something more to look at?
    } ;

#ifdef SEGMENTED_MEMORY
static inline unsigned long
MemoryPool::AllocatorOverhead( unsigned long size )
{
return SegmentOverhead * ( ( size + SegmentSize - 1 ) / SegmentSize ) ;
}

inline
MemoryPool::MemoryPool( unsigned long size )
{
    int 	full_segments	  = int( size / SegmentSize ) ;
    unsigned	last_segment_size = int( size % SegmentSize ) ;

segment_count	   = full_segments + ( last_segment_size > 0 ) ;
segments	   = new char *[ segment_count ] ;
for( int i = 0 ; i < full_segments ; i++ )
    segments[ i ]  = new char[ SegmentSize + sizeof( unsigned ) ] ;
if( last_segment_size > 0 )
    segments[ i ]  = new char[ last_segment_size + sizeof( unsigned ) ] ;
for( i = 0 ; i < segment_count ; i++ ){
    if( segments[ i ] == NULL ){
	cerr << "MemoryPool::MemoryPool(" << size << "): no memory for requested block\n" ;
	exit( -1 ) ;
	}
    segments[ i ] += sizeof( unsigned ) ;
    }
memory_start	  = SegPtr( segments, 0 ) ;
memory_top	  = SegPtr( segments + segment_count - 1, last_segment_size ) ;
free_store_bottom = memory_start ;
free_store_top	  = memory_top ;
}

inline
MemoryPool::~MemoryPool()
{
for( int i = 0 ; i < segment_count ; i++ )
    delete [] (segments[ i ] - sizeof( unsigned )) ;
}
#else

inline unsigned long
MemoryPool::AllocatorOverhead( unsigned long size )
{
return 0 ;
}

inline
MemoryPool::MemoryPool( unsigned long size )
{
if( ( memory_start = new char[ size ] ) == NULL ){
    cerr << "MemoryPool::MemoryPool(" << size << "): no memory for requested block\n" ;
    exit( -1 ) ;
    }
memory_top	  = memory_start + size ;
free_store_bottom = memory_start ;
free_store_top	  = memory_top ;
}

inline
MemoryPool::~MemoryPool()
{
delete [] memory_start ;
}
#endif

inline unsigned long
MemoryPool::GetFreeSpace()
{
return free_store_top - free_store_bottom ;
}

inline void *
MemoryPool::BottomAlloc( int size )
{
    void *	block	   = free_store_bottom ;
    SegPtr	new_bottom = free_store_bottom + size ;

if( new_bottom > free_store_top )
    return NULL ;
#ifdef SEGMENTED_MEMORY
if( new_bottom - free_store_bottom > size ){
    //	Means we run on segmented architecture, and pointer had wrapped
    //	to the next segment. To make MemoryPoolWalker to work correctly,
    //	we will need to update segment limit offset now.
    free_store_bottom.SetLimit() ;
    block = new_bottom - size ;
    }
#endif
free_store_bottom = new_bottom ;
return block ;
}

inline void *
MemoryPool::TopAlloc( int size )
{
    SegPtr	new_top = free_store_top - size ;

if( new_top < free_store_bottom )
    return NULL ;
free_store_top = new_top ;
return (void *) new_top ;
}

inline void
MemoryPool::FlushBottomLine()
{
#ifdef SEGMENTED_MEMORY
free_store_bottom.SetLimit() ; // Allocation chain ended, it's safe to set limit now...
#endif
free_store_bottom = memory_start ;
}

inline
MemoryPoolWalker::MemoryPoolWalker( MemoryPool &pool )
{
memory_ptr   = pool.memory_start ;
memory_limit = pool.free_store_bottom ;
}

inline MemoryPoolWalker &
MemoryPoolWalker::operator += ( int size )
{
memory_ptr  += size ;
#ifdef SEGMENTED_MEMORY
if( memory_ptr.AtLimit() )
    memory_ptr.NextSegment() ;
#endif
return *this ;
}

inline
MemoryPoolWalker::operator void *()
{
return (void *) memory_ptr ;
}

inline
MemoryPoolWalker::operator int()
{
return memory_ptr < memory_limit ;
}
